#!/usr/bin/perl
#
# dbm-server.pl
#
# DBM Server
#
# Copyright 2001 by Anthony Howe.  All rights reserved.
#
# usage: dbm-server.pl [-D][-d|-k][-p port][-h hosts]
#
# -D		Debug.
# -d		Daemon server mode.
# -h hosts	Comma separated list of host names or IPs allowed to connect.
# -k		Kill an already running instance of the server.
# -p port	Listen for UDP packets on this port.
#

### The name of this server.
$myname = 'dbm-server';

### This server runs on a privelaged port number only so that the
### sys.admin takes the necessary care in installing the software.
$PORT = 9866;

### List of hosts allowed to connect to us. This server is intended to be
### a private local service, but we can make allowances.
%HOSTS = ('127.0.0.1' => 1, 'localhost' => 1);

### Where do we keep track of our pid?  Must be an absolute pathname.
$PIDFILE = "/var/run/$myname.pid";

### Absolute pathname of the log file.
$LOGFILE = "/var/log/$myname";

### Location of the DBM databases.
$DBDIR = "/var/spool/$myname";

### How many connections to queue.
$LISTEN = 8;

##########################################################################
#	Nothing to be configured below this point.			 #
##########################################################################
$VERSION = '1.2';
$AUTHOR = 'achowe@snert.com';

use Socket;
use Getopt::Std;
use Sys::Hostname;
use AnyDBM_File;

##########################################################################
#	System Log Routines						 #
##########################################################################

### On some unix systems you have to make sure that syslogd is
### invoked with the -r option to open a network socket for logging.
#use Sys::Syslog;

sub logopen {
	# Separate server log.
	if (defined $_[0]) {
		unless (open(LOG, ">>$_[0]")) {
			die("$myname: couldn't open log \"$log\": $!\n");
		}
		select( (select(LOG), $|=1)[0] );
		chmod 0600, $_[0];
	}

	# Send message to system log.
#	openlog($myname, 'ndelay pid', 'daemon') || logmsg("openlog() failed : $!");
}

sub logmsg {
#	syslog('notice', "$_[0]\n");
	return unless defined LOG;

	my ($s, $m, $h, $day, $mon, $year, $wday, $yday, $isdst) = localtime;

	eval('flock(LOG, 2)');
	printf LOG
		"%d-%02d-%02d %02d:%02d:%02d %s[%d]: ",
		$year+1900, $mon+1, $day, $h, $m, $s, $myname, $$
	;
	print LOG @_, "\n";
	eval('flock(LOG, 8)');
}

sub logclose {
#	closelog();
	close(LOG);
}

sub fatal {
	print STDERR "$myname:", @_, "\n";
	logmsg(@_);
	logclose();
	exit 2;
}

##########################################################################
#	Parse Options							 #
##########################################################################

my $usage = "usage: $myname [-D][-d|-k][-p port][-h hosts]\n";
die($usage) unless getopts('Ddkp:h:');
die($usage) unless @ARGV == 0;

if (defined $opt_D) {
	# Debug paths.
	$PIDFILE = "$myname.pid";
	$LOGFILE = "$myname.log";
	$DBDIR = ".";
}

$PORT = $opt_p if defined $opt_p;

# Allow additional hosts?
foreach (split(/,/, $opt_h)) {
	$HOSTS{$_} = 1;
}

# Start server log.
logopen($LOGFILE);

# Are we already running?
if (-e $PIDFILE) {
	open(PIDFILE); $pid = <PIDFILE>; close(PIDFILE); chomp $pid;
	kill 'TERM', $pid unless $pid == $$ && !$opt_k;
	exit 0 if defined $opt_k;
}

# Create socket server will listen on.
socket(S, PF_INET, SOCK_STREAM, getprotobyname('tcp'));
my $S_addr = sockaddr_in($PORT, INADDR_ANY);
setsockopt(S, SOL_SOCKET, SO_REUSEADDR, 1);

# Put out the "open for business" sign.
fatal("bind failed: $!") unless bind(S, $S_addr);

# The TCP socket equivalent of call-waiting.
fatal("listen failed: $!") unless listen(S, $LISTEN);

# Switch to daemon server mode?
if ($opt_d) {
	$pid = fork;
	exit if $pid;
	fatal("couldn't fork: $!") unless defined $pid;

	# Remove ourself from the controlling terminal and process group
	# that started us.
	use POSIX;
	fatal("can't start new session: $!") unless POSIX::setsid();
	print STDERR "$myname: started [$$]\n";
}

logmsg("server started on port $PORT");

# I'm sure there is a race condition here between where we check
# above, and finally setting the thing here.  Is it important?
open(PIDFILE, ">$PIDFILE") || logmsg("Failed to create \"$PIDFILE\": $!");
print PIDFILE $$, "\n";
close(PIDFILE);

# Convert our possibly canonical name into our real name.
#my $S_host = gethostbyaddr(gethostbyname(hostname()), AF_INET);

##########################################################################
#	Signal Handlers							 #
##########################################################################

$running = 1;

sub signal_handler {
	logmsg("received signal, stopping server.");
	$running = 0;
	close(S);
}

$SIG{INT} = $SIG{TERM} = $SIG{HUP} = \&signal_handler;
$SIG{PIPE} = 'IGNORE';

##########################################################################
#	Main								 #
##########################################################################

srand( time() ^ ($$ + ($$ << 15)) );

# Database name to handle mapping.
my %dbmap = ();

# Fun stuff...
my @sex = (
	"Oh, you are a sad git!",
	"Sorry, you're not my type.",
	"Go play with yourself instead.",
	"Come on, stick it in the floppy slot.",
	"What do you think I am, blonde?",
	"Oh really?!",
	"Sorry, I have a headache.",
	"Take a cold shower.",
	"Go home and sleep it off.",
	"I'm already seeing someone.",
	"I don't believe you love me.",
	"Sex, who needs it.",
	"I know a cute little toaster who's desperate.",
	"Go ask Natalie, I'm busy.",
	"First buy me a drink.",
	"No way! You're 3 beer ugly.",
	"Try the hole in the fence out back.",
	"I hate to imagine what our kids would look like.",
	"Oh lover boy!",
	"You're such a tosser.",
	"First turn out the lights.",
	"But will you respect me in the morning?",
	"Marry me first.",
	"When I'm only a P-120, will you still love me?",
	"You...me...not a chance.",
	"I have better things to do than waste my time with you.",
	"I don't think your man enough.",
	"I don't think your woman enough.",
	"Oh please please go away.",
	"Oh oh oh oh...false alarm.",
	"Ahhhhhh...",
	"Meg Ryan knows how to do it right.",
	"Beg. I love to watch you beg.",
	"I slice, I dice, I make Julien fries.",
	"Yeah baby!",
);

sub respond {
	my ($C, $status, $response) = @_;

	my $msg = pack("CCna*", 0, $status, length($response), $response);

	if (syswrite($C, $msg, length($msg)) < length($msg)) {
		logmsg("send failed $C_host [$C_ip] port $C_port: $!");
	}
}


$DBDIR .= '/' unless $DBDIR =~ m!/$!;
my ($C, $C_port, $C_addr, $C_host, $C_ip, $len);


my ($request, $response, $status, $len);
my ($vn, $command, $dlen, $slen, $klen, $vlen);
my ($secret, $dbname, $key, $value);

# Wait for client connections.
while ($running && ($C = accept(C, S))) {
	($C_port, $C_addr) = sockaddr_in($C);
	$C_host = gethostbyaddr($C_addr, AF_INET);
	$C_ip = inet_ntoa($C_addr);

	unless ($HOSTS{$C_ip} || $HOSTS{$C_host}) {
		logmsg("access denied to $C_host [$C_ip] port $C_port");
		close(C);
		next;
	}
	
	logmsg("received connection from $C_host [$C_ip] port $C_port");

	# Read requests until a quit request, an error, or broken connection.
	while ($running && read(C, $request, 7)) {
		# Parse request header.
		($vn, $command, $dlen, $slen, $klen, $vlen) = unpack("CaCCCna*", $request);

#		logmsg("VN=$vn C=$command, D=$dlen, S=$slen, K=$klen, V=$vlen\n");
		
		# Check version number & header length.
		($status = 0x21), last unless $vn == 0 && length($request) == 7;
			
		# Read remainder of request.
		$len = $dlen + $slen + $klen + $vlen;
		($status = 0x11), last unless read(C, $request, $len);

		# Check length of request.
		($status = 0x22), last unless length($request) == $len;

		# Get fields.
		($dbname, $secret, $key, $value) =
			unpack("a${dlen}a${slen}a${klen}a${vlen}", $request);
		
		($status = 0x01), last if $command eq 'q';

		# Is the DB already open?
		unless (defined $dbmap{$dbname}) {
			$dbmap{$dbname} = ();
			my $db_exist = (-e "$DBDIR$dbname");

			unless (dbmopen(%{ $dbmap{$dbname} }, $DBDIR . $dbname, 0600)) {
				logmsg("failed to open DBM \"$DBDIR$dbname\": $!");
				next;
			}

#			logmsg("opened \"$dbname\"");
			$dbmap{$dbname}{'DATABASE-SECRET'} = $secret unless $db_exist;
		}

		# Do the request and database secrets match?
		($status = 0x23), last unless $secret eq $dbmap{$dbname}{'DATABASE-SECRET'};

		$status = 0x00;
		# Process command.
		if ($command eq 'c') {
#			logmsg("closed \"$dbname\"");
			dbmclose(%{ $dbmap{$dbname} });
			delete $dbmap{$dbname};
			$response = $dbname;
		} elsif ($command eq 'd') {
#			logmsg("delete \"$dbname\" key \"$key\"");
			$response = $dbmap{$dbname}{$key};
			delete $dbmap{$dbname}{$key};
		} elsif ($command eq 'g') {
#			logmsg("get \"$dbname\" key \"$key\"");
			$response = $dbmap{$dbname}{$key};
		} elsif ($command eq 's') {
#			logmsg("set \"$dbname\" key \"$key\"");
			$dbmap{$dbname}{$key} = $value;
			$response = $value;
		} elsif ($command eq 'x') {
			$response = $sex[int rand @sex];
#			logmsg("sex \"$response\"");
			if ($key eq 'TERM') {
				$status = 0x01;
				$running = 0;
				last;
			}
		} else {
			logmsg("unknown command \"$command\"");
			$status = 0x20;
			last;
		}

		respond(C, $status, $response);
	}
	
	respond(C, $status, "END");
	close(C);
}

##########################################################################
#	Clean-Up							 #
##########################################################################

### Bug in ActivePerl for Windows won't handle foreach (keys %dbmap).
#@dbnames = keys %dbmap;
#foreach (@dbnames) {

foreach (keys %dbmap) {
#	logmsg("closed \"$_\"");
	dbmclose(%{ $dbmap{$_} });
}

logmsg("server stopped on port $PORT");
print STDERR "$myname: stopped [$$]\n";
unlink($PIDFILE);
logclose();
close(S);
exit 0;

##########################################################################
#									 #
##########################################################################
__END__